# -*- coding: utf-8 -*-

twitter_base_api_url = "https://{api}.twitter.com/{version}"
twitter_api_version = "1.1"

request_methods = {"get", "post", "put", "delete", "patch", "option", "head"}
streaming_apis = {"stream", "userstream", "sitestream"}

rate_limit_notices = [
    b"Exceeded connection limit for user",
    b"Easy there, Turbo. Too many requests recently. Enhance your calm."
]

formats = [dict(format='PNG'),
           dict(format='JPEG', quality=90, optimize=True)]



# -*- coding: utf-8 -*-

import asyncio
import functools
import io
import json
import os
import sys
import traceback
from urllib.parse import urlparse

import aiohttp

from . import exceptions

try:
    import PIL.Image
except ImportError:
    PIL = None

try:
    import magic
    mime = magic.Magic(mime=True)
except:
    import mimetypes
    mime = mimetypes.MimeTypes()
    magic = None


class JSONObject(dict):
    """
        A dict in which you can access items as attributes

    >>> obj = JSONObject(key=True)
    >>> obj['key'] is obj.key
    True
    """

    def __getattr__(self, key):
        if key in self:
            return self[key]
        raise AttributeError("%s has no property named %s." %
                             (self.__class__.__name__, key))

    def __setattr__(self, *args):
        raise AttributeError("%s instances are read-only." %
                             self.__class__.__name__)
    __delattr__ = __setitem__ = __delitem__ = __setattr__


class PeonyResponse:
    """
        Response objects

    In these object you can access the headers, the request, the url
    and the response
    getting an attribute/item of this object will get the corresponding
    attribute/item of the response

    >>> peonyresponse = PeonyResponse(
    ...     response=JSONObject(key="test"), headers={},
    ...     url="http://google.com", request={}
    ... )
    >>> peonyresponse.key is peonyresponse.response.key  # returns True
    >>>
    >>> peonyresponse = PeonyResponse(
    ...     response=[JSONObject(key="test"), JSONObject(key=1)], headers={},
    ...     url="http://google.com", request={}
    ... )
    >>> # iterate over peonyresponse.response
    >>> for key in peonyresponse:
    ...     pass  # do whatever you want

    Parameters
    ----------
    response : JSONObject or list
        Response object
    headers : dict
        Headers of the response
    url : str
        URL of the request
    request : dict
        Requests arguments
    """

    def __init__(self, response, headers, url, request):
        self.response = response
        self.headers = headers
        self.url = url
        self.request = request

    def __getattr__(self, key):
        """ get attributes from the response """
        return getattr(self.response, key)

    def __getitem__(self, key):
        """ get items from the response """
        return self.response[key]

    def __iter__(self):
        """ iterate over the response """
        return iter(self.response)

    def __str__(self):
        """ use the string of the response """
        return str(self.response)

    def __repr__(self):
        """ use the representation of the response """
        return repr(self.response)

    def __len__(self):
        """ get the lenght of the response """
        return len(self.response)


def error_handler(request):
    """
        The default error_handler

    The decorated request will retry infinitely on any handled error
    The exceptions handled are :class:`asyncio.TimeoutError` and
    :class:`exceptions.RateLimitExceeded`
    """

    @functools.wraps(request)
    async def decorated_request(timeout=10, **kwargs):
        while True:
            try:
                with aiohttp.Timeout(timeout):
                    return await request(**kwargs)

            except exceptions.RateLimitExceeded as e:
                delay = int(e.reset_in) + 1
                fmt = "Sleeping for {}s (rate limit exceeded on endpoint {})"
                print(fmt.format(delay, kwargs['url']), file=sys.stderr)
                await asyncio.sleep(delay)

            except asyncio.TimeoutError:
                fmt = "Request to {url} timed out, retrying"
                print(fmt.format(url=kwargs['url']), file=sys.stderr)

            except:
                raise

    return decorated_request


def get_args(func, skip=0):
    """
        Hackish way to get the arguments of a function

    Parameters
    ----------
    func : function
        Function to get the arguments from
    skip : :obj:`int`, optional
        Arguments to skip, defaults to 0 set it to 1 to skip the
        ``self`` argument of a method.

    Returns
    -------
    tuple
        Function's arguments
    """
    argcount = func.__code__.co_argcount
    return func.__code__.co_varnames[skip:argcount]


def print_error(msg=None, stderr=sys.stderr, error=None):
    """
        Print an exception and its traceback to stderr

    Parameters
    ----------
    msg : :obj:`str`, optional
        A message to add to the error
    stderr : file object
        A file object to write the errors to
    """
    output = [] if msg is None else [msg]
    output.append(traceback.format_exc().strip())

    print(*output, sep='\n', file=stderr)


def loads(json_data, *args, encoding="utf-8", **kwargs):
    """
        Custom loads function with an object_hook and automatic decoding

    Parameters
    ----------
    json_data : str
        The JSON data to decode
    *args
        Positional arguments, passed to :func:`json.loads`
    encoding : :obj:`str`, optional
        The encoding of the bytestring
    **kwargs
        Keyword arguments passed to :func:`json.loads`

    Returns
    -------
    :obj:`dict` or :obj:`list`
        Decoded json data
    """
    if isinstance(json_data, bytes):
        json_data = json_data.decode(encoding)

    return json.loads(json_data, *args, object_hook=JSONObject, **kwargs)


def media_chunks(media, chunk_size, media_size=None):
    """
        read the file by chunks

    Parameters
    ----------
    media : file
        The file object to read
    chunk_size : int
        The size of a chunk in bytes
    media_size : :obj:`int`, optional
        Size of the file in bytes

    Yields
    ------
    bytes
        A chunk of the file
    """
    if media_size is None:
        media_size = get_size(media)

    while media.tell() < media_size:
        yield media.read(chunk_size)


def convert(img, formats):
    """
        Convert the image to all the formats specified

    Parameters
    ----------
    img : PIL.Image.Image
        The image to convert
    formats : list
        List of all the formats to use

    Yields
    ------
    io.BytesIO
        A file object containing the converted image
    """
    for kwargs in formats:
        f = io.BytesIO()
        img.save(f, **kwargs)
        yield f


def optimize_media(file_, max_size, formats):
    """
        Optimize an image

    Resize the picture to the ``max_size``, defaulting to the large
    photo size of Twitter in :meth:`PeonyClient.upload_media` when
    used with the ``optimize_media`` argument.

    Parameters
    ----------
    file_ : file object
        the file object of an image
    max_size : :obj:`tuple` or :obj:`list` of :obj:`int`
        a tuple in the format (width, height) which is maximum size of
        the picture returned by this function
    formats : :obj`list` or :obj:`tuple` of :obj:`dict`
        a list of all the formats to convert the picture to

    Returns
    -------
    file
        The smallest file created in this function
    """
    if not PIL:
        msg = ("Pillow must be installed to optimize a media\n"
               "(pip3 install peony[Pillow])")
        raise RuntimeError(msg)

    img = PIL.Image.open(file_)
    ratio = max(hw / max_hw for hw, max_hw in zip(img.size, max_size))

    # resize the picture (defaults to the 'large' photo size of Twitter
    # in peony.PeonyClient.upload_media)
    if ratio > 1:
        size = tuple(int(hw // ratio) for hw in img.size)
        img = img.resize(size, PIL.Image.ANTIALIAS)

    files = list(convert(img, formats))

    # do not close a file opened by the user
    # only close if a filename was given
    if not hasattr(file_, 'read'):
        img.close()

    files.sort(key=get_size)
    media = files.pop(0)

    for f in files:
        if not f.closed:
            f.close()

    return media


def reset_io(func):
    """
    A decorator to set the pointer of the file to beginning
    of the file before and after the decorated function
    """
    @functools.wraps(func)
    def decorated(media):
        media.seek(0)
        result = func(media)
        media.seek(0)

        return result

    return decorated


@reset_io
def get_media_metadata(media):
    """
        Get the metadata of the file

    Parameters
    ----------
    media : file
        The file to analyze

    Returns
    -------
    str
        The mimetype of the media
    str
        The category of the media on Twitter
    bool
        Tell whether this file is an image or a video
    """
    media_type, media_category = get_type(media)
    is_image = not (media_type.endswith('gif')
                    or media_type.startswith('video'))

    return media_type, media_category, is_image


def get_image_metadata(file_):
    """
        Get all the file's metadata and read any kind of file object

    Parameters
    ----------
    file_ : file object
        A file object of the image

    Returns
    -------
    str
        The mimetype of the media
    str
        The category of the media on Twitter
    bool
        Tell whether this file is an image or a video
    str
        Path to the file
    """
    # try to get the path no matter how the input is
    if isinstance(file_, str):
        path = urlparse(file_).path.strip(" \"'")

        with open(path, 'rb') as original:
            return (*get_media_metadata(original), path)

    elif hasattr(file_, 'read'):
        return (*get_media_metadata(file_), file_)
    else:
        raise TypeError("upload_media input must be a file object or a"
                        "filename")


@reset_io
def get_size(media):
    """
        Get the size of a file

    Parameters
    ----------
    media : file object
        The file object of the media

    Returns
    -------
    int
        The size of the file
    """
    media.seek(0, os.SEEK_END)
    return media.tell()


@reset_io
def get_type(media, path=None):
    """
    Parameters
    ----------
    media : file object
        A file object of the image
    path : str, optional
        The path to the file

    Returns
    -------
    str
        The mimetype of the media
    str
        The category of the media on Twitter
    """
    if magic:
        media_type = mime.from_buffer(media.read(1024))
    else:
        media_type = None
        if path:
            media_type = mime.guess_type(path)[0]

        if media_type is None:
            msg = ("Could not guess the mimetype of the media.\n"
                   "Please consider installing python-magic\n"
                   "(pip3 install peony-twitter[magic])")
            raise RuntimeError(msg)

    if media_type.startswith('video'):
        media_category = "tweet_video"
    elif media_type.endswith('gif'):
        media_category = "tweet_gif"
    else:
        media_category = "tweet_image"

    return media_type, media_category



# -*- coding: utf-8 -*-
"""
Peony Clients

:class:`BasePeonyClient` only handles requests while
:class:`PeonyClient` adds some methods that could help when using
the Twitter APIs, with a method to upload a media
"""

import asyncio
import io

import aiohttp

from . import exceptions, general, oauth, utils
from .api import APIPath, StreamingAPIPath
from .commands import EventStreams, task
from .stream import StreamContext
from .oauth import OAuth1Headers


class BasePeonyClient:
    """
        Attributes/items become a :class:`api.APIPath` or
        :class:`api.StreamingAPIPath` automatically

    This class only handles the requests and makes accessing Twitter's
    APIs easy.

    Parameters
    ----------
    streaming_apis : iterable, optional
        Iterable containing the streaming APIs subdomains
    base_url : :obj:`str`, optional
        Format of the url for all the requests
    api_version : :obj:`str`, optional
        Default API version
    suffix : :obj:`str`, optional
        Default suffix of API endpoints
    loads : :obj:`function`, optional
        Function used to load JSON data
    error_handler : :obj:`function`, optional
        Requests decorator
    session : :obj:`asyncio.ClientSession`, optional
        session to use to make requests
    proxy : str
        proxy used with every request
    loop : event loop, optional
        An event loop, if not specified :func:`asyncio.get_event_loop`
        is called
    """

    def __init__(self, consumer_key, consumer_secret,
                 access_token=None,
                 access_token_secret=None,
                 bearer_token=None,
                 auth=None,
                 headers=None,
                 streaming_apis=None,
                 base_url=None,
                 api_version=None,
                 suffix='.json',
                 loads=utils.loads,
                 error_handler=utils.error_handler,
                 session=None,
                 proxy=None,
                 loop=None,
                 **kwargs):

        if streaming_apis is None:
            self.streaming_apis = general.streaming_apis
        else:
            self.streaming_apis = streaming_apis

        if base_url is None:
            self.base_url = general.twitter_base_api_url
        else:
            self.base_url = base_url

        if api_version is None:
            self.api_version = general.twitter_api_version
        else:
            self.api_version = api_version

        if auth is None:
            auth = OAuth1Headers

        if headers is None:
            headers = {}

        self.proxy = proxy

        self._suffix = suffix

        self._loads = loads
        self.error_handler = error_handler

        self.loop = asyncio.get_event_loop() if loop is None else loop

        self._session = session

        # all the possible args required by headers in :mod:`peony.oauth`
        kwargs = {
            'consumer_key': consumer_key,
            'consumer_secret': consumer_secret,
            'access_token': access_token,
            'access_token_secret': access_token_secret,
            'bearer_token': bearer_token,
            'client': self
        }

        # get the args needed by the auth parameter on initialization
        args = utils.get_args(auth.__init__, skip=1)

        # keep only the arguments required by auth on init
        kwargs = {key: value for key, value in kwargs.items()
                  if key in args}

        self.headers = auth(**kwargs, **headers)

        self.__setup = {'event': asyncio.Event(),
                        'state': False}

        if not self.loop.is_running():
            self.loop.run_until_complete(self.setup())

    async def setup(self):
        """
            set up the client on the first request
        """
        if not self.__setup['state']:
            self.__setup['state'] = True

            if self._session is None:
                self._session = aiohttp.ClientSession()

            prepare_headers = self.headers.prepare_headers()
            if prepare_headers is not None:
                await prepare_headers

            self.__setup['event'].set()

            if callable(self.init_tasks):
                init_tasks = self.init_tasks()
            else:
                init_tasks = self.init_tasks

            if init_tasks is not None:
                await asyncio.wait(init_tasks)

        await self.__setup['event'].wait()

    def init_tasks(self):
        """ tasks executed on initialization """
        pass

    def __getitem__(self, values):
        """
            Access the api you want

        This permits the use of any API you could know about

        For most api you only need to type

        >>> self[api]  # api is the api you want to access

        You can specify a custom api version using the syntax

        >>> self[api, version]  # version is the api version as a str

        For more complex requests

        >>> self[api, version, suffix, base_url]

        Returns
        -------
        api.BaseAPIPath
            To access an API endpoint
        """
        defaults = (None, self.api_version, self._suffix, self.base_url)
        keys = ['api', 'version', 'suffix', 'base_url']

        if isinstance(values, dict):
            # set values in the right order
            values = [values[key] for key in keys]
        elif isinstance(values, set):
            raise TypeError('Cannot use a set to access an api, '
                            'please use a dict, a tuple or a list instead')
        elif isinstance(values, str):
            values = [values, *defaults[1:]]
        elif values:
            if len(values) < len(keys):
                padding = (None,) * (len(keys) - len(values))
                values += padding

            values = [default if value is None else value
                      for value, default in zip(values, defaults)
                      if (value, default) != (None, None)]

        api, version, suffix, base_url = values

        base_url = base_url.format(api=api, version=version).rstrip('/')

        # use StreamingAPIPath if subdomain is in self.streaming_apis
        if api in self.streaming_apis:
            return StreamingAPIPath([base_url], suffix=suffix, client=self)
        else:
            return APIPath([base_url], suffix=suffix, client=self)

    def __getattr__(self, api):
        """
            Access the api you want

        Same as calling client[api]

        Returns
        -------
        api.BaseAPIPath
            To access an API endpoint
        """
        return self[api]

    def __del__(self):
        self._session.close()

    async def request(self, method, url,
                      headers=None,
                      json=False,
                      session=None,
                      **kwargs):
        """
            Make requests to the REST API

        Parameters
        ----------
        method : str
            Method to be used by the request
        url : str
            URL of the ressource
        headers : peony.oauth.PeonyHeaders
            Custom headers (doesn't overwrite `Authorization` headers)
        session : :obj:`aiohttp.ClientSession`, optional
            Client session used to make the request
        json : bool
            Force json decoding

        Returns
        -------
        utils.PeonyResponse
            Response to the request
        """
        await self.setup()

        # prepare request arguments, particularly the headers
        req_kwargs = self.headers.prepare_request(
            method=method,
            url=url,
            headers=headers,
            **kwargs
        )

        if 'proxy' not in req_kwargs:
            req_kwargs['proxy'] = self.proxy

        session = session if (session is not None) else self._session

        # make the request
        async with session.request(**req_kwargs) as response:
            if response.status // 100 == 2:
                if json or url.endswith(".json") and json is not None:
                    # decode as json
                    content = await response.json(loads=self._loads)
                else:
                    # decode as text
                    content = await response.text()

                return utils.PeonyResponse(
                    response=content,
                    headers=response.headers,
                    url=response.url,
                    request=req_kwargs
                )
            else:  # throw exception if status is not 2xx
                await exceptions.throw(response)

    def stream_request(self, method, url, headers=None, _session=None,
                       *args, **kwargs):
        """
            Make requests to the Streaming API

        Parameters
        ----------
        method : str
            Method to be used by the request
        url : str
            URL of the ressource
        headers : dict
            Custom headers (doesn't overwrite `Authorization` headers)
        _session : :obj:`aiohttp.ClientSession`, optional
            The session to use for this specific request, the session
            given as argument of :meth:`__init__` is used by default

        Returns
        -------
        stream.StreamContext
            Stream context for the request
        """
        return StreamContext(
            method, url, self,
            *args,
            headers=headers,
            session=_session,
            **kwargs
        )


class PeonyClient(BasePeonyClient):
    """
        A client with an easy handling of tasks

    You can create tasks by decorating a function from a child
    class with :class:`peony.task`

    You also attach a :class:`EventStream` to a subclass using
    the :func:`event_stream` of the subclass

    After creating an instance of the child class you will be able
    to run all the tasks easily by executing :func:`get_tasks`
    """

    def init_tasks(self):
        tasks = [self.__get_twitter_configuration()]
        if isinstance(self.headers, oauth.OAuth1Headers):
            tasks.append(self.__get_user())
        return tasks

    async def __get_twitter_configuration(self):
        """
        create a ``twitter_configuration`` attribute with the response
        of the endpoint
        https://api.twitter.com/1.1/help/configuration.json
        """
        api = self['api', general.twitter_api_version,
                   ".json", general.twitter_base_api_url]

        req = api.help.configuration.get()
        self.twitter_configuration = await req

    async def __get_user(self):
        """
        create a ``user`` attribute with the response of the endpoint
        https://api.twitter.com/1.1/account/verify_credentials.json
        """
        api = self['api', general.twitter_api_version,
                   ".json", general.twitter_base_api_url]

        req = api.account.verify_credentials.get()
        self.user = await req

    async def _chunked_upload(self, media,
                              path=None,
                              media_type=None,
                              media_category=None,
                              chunk_size=2**20,
                              **params):
        """
            upload media in chunks

        Parameters
        ----------
        media : file object
            a file object of the media
        path : :obj:`str`, optional
            filename of the media
        media_type : :obj:`str`, optional
            mime type of the media
        media_category : :obj:`str`, optional
            twitter media category
        chunk_size : :obj:`int`, optional
            size of a chunk in bytes
        params : :obj:`dict`, optional
            additional parameters of the request

        Returns
        -------
        utils.PeonyResponse
            Response of the request
        """
        media_size = utils.get_size(media)

        if media_type is None or media_category is None:
            media_type, media_category = utils.get_type(media, path)

        response = await self.upload.media.upload.post(
            command="INIT",
            total_bytes=media_size,
            media_type=media_type,
            media_category=media_category,
            **params
        )

        media_id = response['media_id']

        chunks = utils.media_chunks(media, chunk_size, media_size)

        for i, chunk in enumerate(chunks):
            await self.upload.media.upload.post(command="APPEND",
                                                media_id=media_id,
                                                media=chunk,
                                                segment_index=i,
                                                _json=None)

        status = await self.upload.media.upload.post(command="FINALIZE",
                                                     media_id=media_id)

        if 'processing_info' in status:
            while status['processing_info']['state'] != "succeeded":
                if status['processing_info'].get('state', "") == "failed":
                    error = status['processing_info'].get('error', {})

                    message = error.get('message', str(status))

                    raise exceptions.MediaProcessingError(data=status,
                                                          message=message,
                                                          **params)

                delay = status['processing_info']['check_after_secs']
                await asyncio.sleep(delay)
                status = await self.upload.media.upload.get(
                    command="STATUS",
                    media_id=media_id,
                    **params
                )

        return response

    async def upload_media(self, file_,
                           auto_convert=False,
                           formats=None,
                           max_size=None,
                           chunked=False,
                           size_limit=None,
                           **params):
        """
            upload a media on twitter

        Parameters
        ----------
        file_ : :obj:`str` or file
            Path to the file or file object
        auto_convert : :obj:`bool`, optional
            If set to True the media will be optimized by calling
            :func:`utils.optimize_media`
        formats : :obj:`list`, optional
            A list of all the formats to try to optimize the media
        max_size : :obj:`tuple`, optional
            Max size of the picture in the (width, height) format
        chunked : :obj:`bool`, optional
            If True, force the use of the chunked upload for the media
        size_limit : :obj:`int`, optional
            If set, the media will be sent using a multipart upload if
            its size is over ``sizelimit`` bytes

        Returns
        -------
        utils.PeonyResponse
            response of the request
        """
        formats = formats or general.formats

        image_metadata = utils.get_image_metadata(file_)
        media_type, media_category, is_image, file_ = image_metadata

        if is_image and auto_convert:
            if not max_size:
                photo_sizes = self.twitter_configuration['photo_sizes']
                large_sizes = photo_sizes['large']
                max_size = large_sizes['w'], large_sizes['h']

            media = utils.optimize_media(file_, max_size, formats)
        elif hasattr(file_, 'read'):
            file_.seek(0)
            media = io.BytesIO(file_.read())
            file_.seek(0)
        else:
            media = open(file_, 'rb')

        if not isinstance(self.twitter_configuration, APIPath) or size_limit:
            if size_limit is None:
                size_limit = self.twitter_configuration['photo_size_limit']
            size_test = utils.get_size(media) > size_limit
        else:
            size_test = False

        if size_test or chunked:
            args = media, file_, media_type, media_category
            response = await self._chunked_upload(*args, **params)
        else:
            response = await self.upload.media.upload.post(media=media,
                                                           **params)

        if not media.closed:
            media.close()

        return response

    @classmethod
    def event_stream(cls, event_stream):
        """ Decorator to attach an event stream to the class """
        if getattr(cls, '_streams', None) is None:
            cls._streams = EventStreams()

        cls._streams.append(event_stream)
        return event_stream

    def get_tasks(self):
        """
            Get the tasks attached to the instance

        Returns
        -------
        list
            List of tasks (:class:`asyncio.Task`)
        """
        funcs = [getattr(self, key) for key in dir(self)]
        tasks = [self.loop.create_task(func(self))
                 for func in funcs if isinstance(func, task)]

        if isinstance(self._streams, EventStreams):
            tasks.extend(self._streams.get_tasks(self))

        return tasks

    async def run_tasks(self):
        """ Run the tasks attached to the instance """
        await self.setup()
        await asyncio.wait(self.get_tasks())

    def run(self):
        """ Run the tasks attached to the instance """
        self.loop.create_task(self.run_tasks())
        try:
            self.loop.run_forever()
        except KeyboardInterrupt:
            pending = asyncio.Task.all_tasks(loop=self.loop)
            gathered = asyncio.gather(*pending, loop=self.loop)
            try:
                gathered.cancel()
                self.loop.run_until_complete(gathered)

                # we want to retrieve any exceptions to make sure that
                # they don't nag us about it being un-retrieved.
                gathered.exception()
            except:
                pass
        finally:
            self.loop.close()



# -*- coding: utf-8 -*-

import re
from functools import wraps

from . import utils
from ..utils import print_error


def process_keys(func):
    """
    Raise error for keys that are not strings
    and add the prefix if it is missing
    """

    @wraps(func)
    def decorated(self, k, *args):
        if not isinstance(k, str):
            msg = "%s: key must be a string" % self.__class__.__name__
            raise ValueError(msg)

        if not k.startswith(self.prefix):
            k = self.prefix + k

        return func(self, k, *args)

    return decorated


class Functions(dict):
    """
        Functions of an event handler

    When a key is found in the text given as an argument to
    :meth:`_run` the corresponding function is called.

    Parameters
    ----------
    prefix : str
        Prefix of the functions
    strict : bool
        In strict mode the command must be at the beginning of the
        string
    args
        Positional arguments passed to :func:`dict.__init__`
        for those who want to do some shady things
    kwargs
        Keyword arguments passed to :func:`dict.__init__`
    """

    def __init__(self, *args, prefix=None, strict=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.prefix = prefix
        ex = re.escape(prefix) + r"[^\s]+"
        self.prog = re.compile(ex)
        self.strict = strict

    @process_keys
    def __getitem__(self, k):
        """ get the function you want """
        return super().__getitem__(k)

    @process_keys
    def __setitem__(self, k, v):
        if self.prefix is None:
            raise RuntimeError("To add commands you must set a prefix")

        super().__setitem__(k, v)

    def __repr__(self):
        return "{classname}({super})".format(
            classname=self.__class__.__name__,
            super=", ".join(super().keys())
        )

    def _get(self, text):
        """
            Analyze the text to get the right function

        Parameters
        ----------
        text : str
            The text that could call a function
        """
        if self.strict:
            match = self.prog.match(text)
            if match:
                cmd = match.group()
                if cmd in self:
                    return cmd
        else:
            words = self.prog.findall(text)
            for word in words:
                if word in self:
                    return word

    async def run(self, *args, data):
        """ run the function you want """
        cmd = self._get(data.text)

        try:
            if cmd is not None:
                command = self[cmd](*args, data=data)
                return await utils.execute(command)

        except:
            fmt = "Error occurred while running function {cmd}:"
            print_error(fmt.format(cmd=cmd))

    def __call__(self, func, name=None):
        name = name or func.__name__
        self[name] = func

        return func


class Commands(Functions):

    def __init__(self, prefix=None):
        super().__init__(prefix=prefix)

        @self
        def help(_self, data, *args, **kwargs):
            """ show commands help """
            kdoc = [
                (key, utils.doc(value))
                for key, value in self.items()
                if utils.permission_check(data,
                                          _permissions=_self.permissions,
                                          command=value)
            ]

            kdoc.sort(key=self._key)

            msg = ["{key}: {doc}".format(key=key, doc=doc)
                   for key, doc in kdoc]

            return "\n".join(msg)

    def _key(self, item):
        key, __ = item
        if key == self.prefix + "help":
            return ''   # /help is the first in the message
        else:
            return key  # sort other commands alphabeticaly

    def restricted(self, *permissions):
        def decorator(func):
            @wraps(func)
            async def decorated(_self, *args, data):
                permission = utils.permission_check(
                    data,
                    _permissions=_self.permissions,
                    permissions=permissions
                )

                if permission:
                    argcount = func.__code__.co_argcount - 1
                    args = (*args, data)[:argcount]
                    cmd = func(_self, *args)

                    return await utils.execute(cmd)

            decorated.permissions = permissions

            return self(decorated)

        return decorator



# -*- coding: utf-8 -*-


class BaseIterator:
    """
        Asynchronous iterator

    Parameters
    ----------
    _request
        Main request
    **kwargs
        Parameters of the request
    """

    def __init__(self, _request, **kwargs):
        self.request = _request
        self.kwargs = kwargs

    def __aiter__(self):
        return self

    async def __anext__(self):
        pass


class IdIterator(BaseIterator):
    """
        Iterate using ids

    It is the parent class of MaxIdIterator and SinceIdIterator

    Parameters
    ----------
    _request :
        Main request
    _parameter : str
        Parameter to change for each request
    _i : int
        Index of the value for the next request
    _force : bool
        Keep the iterator after empty responses
    kwargs:
        Request parameters
    """

    def __init__(self, _request, _parameter, _i, _force=False, **kwargs):
        """ Keep all the arguments as class attributes """
        self.param = _parameter
        self.force = _force
        self.i = _i
        super().__init__(_request, **kwargs)

    async def __anext__(self):
        """ return each response until getting an empty response """
        response = await self.request(**self.kwargs)

        if response:
            response = await self.call_on_response(response)
        elif not self.force:
            raise StopAsyncIteration

        return response

    async def call_on_response(self, response):
        """
            The parameter is set to the id of the tweet at index i - 1
        """
        self.kwargs[self.param] = response[self.i]['id'] - 1
        return response


class MaxIdIterator(IdIterator):
    """
        Iterator for endpoints using max_id

    Parameters
    ----------
    _request:
        Main request
    kwargs:
        Parameters of the request
    """

    def __init__(self, _request, **kwargs):
        super().__init__(_request,
                         _parameter="max_id",
                         _i=-1,
                         _force=False,
                         **kwargs)


class SinceIdIterator(IdIterator):
    """
        Iterator for endpoints using since_id

    Parameters
    ----------
    _request
        Main request
    _force : bool
        Keep the iterator after empty responses
    _fill_gaps : bool
        Fill the gaps (if there are more than ``count`` tweets to get)
    **kwargs
        Parameters of the request
    """

    def __init__(self, _request, _force=True, _fill_gaps=True, **kwargs):
        super().__init__(_request,
                         _parameter="since_id",
                         _i=0,
                         _force=_force,
                         **kwargs)

        self.fill_gaps = _fill_gaps

    async def set_param(self, response):
        if self.fill_gaps:
            self.kwargs[self.param] = response[self.i]['id'] - 1
        else:
            self.kwargs[self.param] = response[self.i]['id']

    async def call_on_response(self, response):
        """
        Try to fill the gaps and strip last tweet from the response
        if its id is that of the first tweet of the former response

        Parameters
        ----------
        response : dict or list
            The response
        """
        since_id = self.kwargs.get(self.param, 0)
        if self.fill_gaps:
            since_id += 1

        if since_id != response[-1]['id']:
            if self.fill_gaps:
                responses = with_max_id(
                    _request=self.request,
                    max_id=response[-1]['id'] - 1,
                    **self.kwargs
                )

                async for tweets in responses:
                    response.extend(tweets)

            if since_id != response[-1]['id']:
                await self.set_param(response)
                return response

        await self.set_param(response)
        return response[:-1]


class CursorIterator(BaseIterator):
    """
        Iterate using a cursor

    Parameters
    ----------
    _request
        Main request
    **kwargs
        Parameters of the request
    """

    async def __anext__(self):
        """ return each response until getting 0 as next cursor """
        if self.kwargs.get('cursor', -1) != 0:
            response = await self.request(**self.kwargs)
            self.kwargs['cursor'] = response['next_cursor']
            return response
        else:
            raise StopAsyncIteration


with_max_id = MaxIdIterator
with_since_id = SinceIdIterator
with_cursor = CursorIterator



# -*- coding: utf-8 -*-

from functools import wraps

from .event_handlers import EventHandler


def get_value(func):
    value = func()

    if value is None:
        value = (func.__name__,)
    elif isinstance(value, str):
        value = (value,)

    return value


class Handler:

    def __init__(self, value):
        self.value = value

    def __call__(self, func, *args, **kwargs):
        return EventHandler(*self.value, *args, func=func, **kwargs)

    def with_prefix(self, prefix):

        def decorated(func):
            return self(func, prefix=prefix)

        return decorated


class Event(list):

    def __init__(self, value):
        super().__init__(value)
        self.handler = Handler(value)


class Events(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aliases = {}

    def __getattr__(self, key):
        return self[key]

    def __setitem__(self, key, value):
        super().__setitem__(key, Event(value))

    def alias(self, *keys):

        def decorator(func):
            value = get_value(func) if callable(func) else func
            self(func, value)

            for key in keys:
                if "{name}" in key:
                    key = key.format(name=func.__name__)

                self[key] = value
                self.aliases[key] = self[func.__name__]

            return func

        return decorator

    @staticmethod
    def event(func):
        value = get_value(func)

        @wraps(func)
        def decorated():
            return [('event', value)]

        return decorated

    @property
    def no_aliases(self):
        return {key: value for key, value in self.items()
                if key not in self.aliases}

    def __call__(self, func, value=None):
        value = value or get_value(func)
        self[func.__name__] = value

        return func

events = Events()

on = 'on_{name}'


@events.alias('on_connect', 'connect')
def friends():
    pass


@events.alias('on_dm', on)
def direct_message():
    pass


@events.alias(on)
def tweet():
    return 'text'


@events.alias(on, 'deleted_tweet')
def delete():
    pass


@events.alias('location_deleted')
def scrub_geo():
    pass


@events
def limit():
    pass


@events
def status_withheld():
    pass


@events
def user_withheld():
    pass


@events.alias(on)
def disconnect():
    pass


@events.alias(on, 'stall_warning')
def warning():
    pass


@events.alias(on)
@events.event
def user_update():
    pass


@events.alias(on)
@events.event
def follow():
    pass


@events.alias(on)
@events.event
def unfollow():
    pass


@events.alias(on)
def access_revoked():
    pass


@events.alias(on)
@events.event
def block():
    pass


@events.alias(on)
@events.event
def unblock():
    pass


@events.alias(on)
@events.event
def favorite():
    pass


@events.alias(on)
@events.event
def unfavorite():
    pass


@events.alias(on)
@events.event
def list_created():
    pass


@events.alias(on)
@events.event
def list_destroyed():
    pass


@events.alias(on)
@events.event
def list_updated():
    pass


@events.alias(on)
@events.event
def list_member_added():
    pass


@events.alias(on)
@events.event
def list_member_removed():
    pass


@events.alias(on)
@events.event
def list_user_subscribed():
    pass


@events.alias(on)
@events.event
def list_user_unsubscribed():
    pass


@events.alias(on)
@events.event
def quoted_tweet():
    pass


@events.alias('control_message')
@events
def control():
    pass


@events.alias(on, 'on_restart', 'restart')
@events
def stream_restart():
    pass


@events.alias(on, 'reconnect', 'on_reconnect')
@events
def reconnecting_in():
    pass


@events.alias('sitestream_event')
@events
def envelope():
    def decorated(*args, **kwargs):
        return ('for_user', ('message', [*args, *kwargs.items()]))
    return
